"""
This module provides the core of SPADE, namely the spade2d function.
"""

import numpy as np


class SpadeCandidates:
    """
    Candidates list for SPADE detected objects.
    """

    def __init__(self, max_candidates, shapes, image_shape):
        self.list = np.zeros((max_candidates, 3), np.uint32)
        self.scores = np.zeros(max_candidates, np.float64)
        self.idx = np.array(0, np.int64)
        self.y, self.x = image_shape
        self.shapes = shapes
        self.labeled_image = np.zeros(
            (self.y + shapes.image_extension_size * 2,
             self.x + shapes.image_extension_size * 2),
            dtype=np.uint16)

    def get_list(self):
        return self.list[:self.idx]

    def get_scores(self):
        return self.scores[:self.idx]

    # TODO: handle RuntimeWarnings raised here
    def add(self, scores, threshold, x, y, possible_shapes_idx):
        valid_scores = np.where(scores >= threshold)
        n_candidates = len(valid_scores[0])
        if n_candidates:
            self.list[self.idx:self.idx + n_candidates, 0] = y
            self.list[self.idx:self.idx + n_candidates, 1] = x
            self.list[self.idx:self.idx + n_candidates, 2] = \
                possible_shapes_idx[0][valid_scores]
            self.scores[self.idx:self.idx + n_candidates] = \
                scores[valid_scores].ravel()
            self.idx += n_candidates

    def sort(self):
        self.list = \
            self.list[:self.idx][self.scores[:self.idx].argsort()][::-1]

    def eliminate_overlapping(self, object_idx_start=1, update_list=False,
                              max_objects=float('Inf')):
        i = 0
        ext_size = self.shapes.image_extension_size
        one = np.array(1, np.uint16)
        # obj number to fill label matrix with
        object_idx = np.array(object_idx_start, np.uint16)
        number_of_candidates = len(self.list)
        object_indices = np.zeros(number_of_candidates, dtype=np.uint32)
        while i < number_of_candidates:
            y, x, shape_id = self.list[i, :]
            local_labeled_image = self.labeled_image[
                                  y:y + self.shapes.grid_size,
                                  x:x + self.shapes.grid_size]
            # if no overlap...
            if not local_labeled_image[self.shapes[shape_id]].any():
                local_labeled_image += self.shapes[shape_id] * object_idx
                object_indices[object_idx - object_idx_start] = i
                object_idx += one
                if object_idx >= max_objects:
                    break
            i += 1
        self.labeled_image = self.labeled_image[ext_size:-ext_size,
                        ext_size:-ext_size]
        if update_list:
            # If return_cand=True, add a last column indicating object number
            # in the labeled image.
            self.list = np.hstack((
                self.list[object_indices[:object_idx - object_idx_start]] -
                    [ext_size, ext_size, 0],
                np.arange(object_idx_start, object_idx)[:, np.newaxis]))


# TODO: write a long and detailed docstring here
def spade2d(image, shapes_library, threshold, data_binding_function=None,
            mask=None, potential_centers=None, return_candidates_list=False,
            object_idx_start=1, max_candidates=10000000):
    """
    Main SPADE function.
    """
    # If no mask or potential centers are passed, creates bool arrays filled
    # with True, same dimensions as image, i.e. inspect all pixels.
    if mask is None:
        mask = np.ones_like(image, bool)
    if potential_centers is None:
        potential_centers = mask
    # Preallocate memory for potential candidates list.
    candidates = SpadeCandidates(max_candidates,
                                 shapes_library,
                                 image.shape)
    # We need to extend the image because it is possible for a object 
    # coordinates to be negative because the shapes coordinates is the top
    # left corner.
    extended_image = np.pad(image.astype(np.float64),
                            shapes_library.image_extension_size,
                            'constant')
    extended_mask = np.pad(mask, shapes_library.image_extension_size,
                           'constant')
    inverted_mask = np.logical_not(extended_mask)
    extended_potential_centers = np.pad(potential_centers,
                                        shapes_library.image_extension_size,
                                        'constant')
    for centered_y, centered_x in np.argwhere(extended_potential_centers):
        real_y = centered_y - shapes_library.half_shape_size
        real_x = centered_x - shapes_library.half_shape_size
        shape_window = (slice(real_y,
                              real_y + shapes_library.grid_size),
                        slice(real_x,
                              real_x + shapes_library.grid_size)
                        )
        local_inverted_mask = inverted_mask[shape_window]
        possible_shapes_bool = np.logical_not(
            (shapes_library & local_inverted_mask).any(axis=(1, 2)))
        if possible_shapes_bool.any():
            possible_shapes_idx = np.where(possible_shapes_bool)
            possible_shapes = shapes_library[possible_shapes_idx]
            ring_window = (slice(real_y - 1,
                                 real_y + shapes_library.grid_size + 1),
                           slice(real_x - 1,
                                 real_x + shapes_library.grid_size + 1)
                           )
            ring_mask = extended_mask[ring_window]
            masked_rings = shapes_library.rings[possible_shapes_idx] & \
                           ring_mask
            if data_binding_function is None:
                # If no data binding function is passed to the detector,
                # use the faster equivalent of mean_difference.
                rings_means = dot_means(masked_rings,
                                        extended_image[ring_window])
                shapes_means = dot_means(possible_shapes,
                                         extended_image[shape_window])
                scores = shapes_means - rings_means
            else:
                shapes_pixels = np.where(shapes_library[possible_shapes_idx],
                                         extended_image[shape_window],
                                         np.nan)
                rings_pixels = np.where(masked_rings,
                                        extended_image[ring_window],
                                        np.nan)
                scores = data_binding_function(shapes_pixels, rings_pixels)
            candidates.add(scores, threshold, real_x, real_y,
                           possible_shapes_idx)
    # Sort candidates by scores.
    candidates.sort()
    candidates.eliminate_overlapping(
        object_idx_start=object_idx_start,
        update_list=True if return_candidates_list else False)

    if return_candidates_list:
        return candidates.labeled_image, candidates.list
    else:
        return candidates.labeled_image


def dot_means(mask, array, ax=(1, 2)):
    # Fast way to return an array of masked (by another array) means.
    # Thanks Divakar from stack overflow for the idea.
    counts = mask.sum(axis=ax)
    # TODO: handle RuntimeWarnings raised here
    return np.dot(mask.reshape(mask.shape[0], -1), array.ravel()) / counts



from skimage.io import imsave
import numpy as np


def random_bright_colors(threshold=0.5):
    while True:
        rgb = np.array([0, 0, 0])
        while rgb.sum() < threshold:
            rgb = np.random.random(3)
        yield rgb


def imsave3d_overlay(file_name, grayscale_image3d, mask, labeled_image,
                     background_color=(0.1, 0, 0.1),
                     colors=random_bright_colors()):
    maximum_intensity = grayscale_image3d.max()
    rgb_image = np.array([(grayscale_image3d / maximum_intensity)] * 3) \
        .transpose(1, 2, 3, 0).astype(np.float16)
    original_rgb_image = np.copy(rgb_image)
    rgb_image[np.logical_not(mask)] = background_color
    for object_idx in range(1, labeled_image.max() + 1):
        rgb_image[labeled_image == object_idx] = next(colors)
    imsave(file_name, (np.concatenate((rgb_image, original_rgb_image),
                                     axis=2)*255).astype(np.uint8))



from skimage.io import imread, imshow
from skimage.filters import threshold_otsu
from skimage.color import label2rgb
from numpy import percentile
import matplotlib.pyplot as plt

from spade.detection_2d import spade2d
from spade.shapes.library import potatoids_5x5__smallest_4_pix


# Load the example image.
image = imread("example_2d.png")

# Separate cell image from background, using by Otsu's thresholding method.
cell = image > threshold_otsu(image)

# Focus on brightest pixels only
potential_centers = image > percentile(image[cell], 99)

# Detect particles.
particles = spade2d(image=image,
                    shapes_library=potatoids_5x5__smallest_4_pix,
                    threshold=20,
                    potential_centers=potential_centers,
                    mask=cell)

# Show detected particles as overlay on our original image.
imshow(label2rgb(particles, image, bg_label=0))
plt.show()



#!/usr/bin/env python

from argparse import ArgumentParser
from sys import exit
from re import sub

from skimage.io import imread, imsave
from skimage.filters import threshold_otsu
from numpy import percentile, ones_like, array_str

from .detection_2d import spade2d


args = ArgumentParser(description='Command-line utility for SPADE.\n'
                                  'It is not recommended to use this '
                                  'interface for batch image processing as '
                                  'each process initialization is slow. Use '
                                  'a python script instead.')
args.add_argument('INPUT_FILE', type=str, help='The image to process. If 3D,'
                                               'the maximum intensity '
                                               'projection will be used.')
args.add_argument('THRESHOLD', type=float, help='The SPADE threshold to use.')
args.add_argument('--compute-mask', '-c',
                  dest='compute_mask',
                  action='store_true',
                  default=False,
                  help='Use Otsu\'s thresholding method to compute the image '
                       'mask.\n Ignored if a mask file is specified.')
args.add_argument('--normalize-mean', '-n',
                  dest='normalize_mean',
                  action='store_true',
                  help='Normalize image mean before detection.',
                  default=False)
args.add_argument('--output', '-o', type=str, default='/tmp/spade_result.png',
                  help='The output image. Each particle will have a '
                       'different grayscale value.\n'
                       '/tmp/spade_result.png by default.')
args.add_argument('--data-binding-function', '-f',
                  dest='data_binding_function',
                  type=str,
                  default=None,
                  help='The data binding function to use.\n'
                       'mean_difference by default. Possible choices: '
                       'max_min_difference, median_difference, '
                       'bhattacharryya_distance, '
                       'pseudo_bhattacharryya_distance and t_test')
args.add_argument('--mask', '-m',
                  type=str,
                  default=None,
                  help='An image used as a mask for the detection.\n'
                       'Any pixels above zero will be interpreted as '
                       'positive.')
args.add_argument('--shapes-library', '-s',
                  dest='shapes_library',
                  type=str,
                  default='potatoids_5x5__smallest_3_pix',
                  help='The shapes library to use.\n'
                       'Default is potatoids_5x5__smallest_3_pix. Possibles '
                       'choices: potatoids_5x5__smallest_1_pix to ...4_pix,'
                       'potatoids_3x3_smallest_1_pix to ...4_pix')
args.add_argument('--restrict', '-r',
                  type=float,
                  default=0,
                  help='Restrict the shapes potential centers to the n '
                       'percent brightest pixels.')
args = args.parse_args()


if args.data_binding_function is not None:
    try:
        exec('from spade.data_binding import {} '
             'as data_binding_function'.format(args.data_binding_function))
    except ImportError:
        print('Cannot find data binding function {}.'.format(
            args.data_binding_function))
        exit(1)
else:
    data_binding_function = None

try:
    exec('from spade.shapes.library import {} as shapes_library'.format(
        args.shapes_library))
except ImportError:
    print('Cannot find shapes library {}.'.format(args.shapes_library))
    exit(1)
print('Using shapes library {}.'.format(args.shapes_library))

try:
    image = imread(args.INPUT_FILE)
except FileNotFoundError:
    print('Cannot find image file {}.'.format(args.INPUT_FILE))
    exit(1)

if image.ndim > 2:
    print('3D image passed as input, using maximum intensity projection.')
    image = image.max(axis=0)

if args.mask is None:
    if args.compute_mask:
        print('Computing mask using Otsu\'s thresholding.')
        mask = image > threshold_otsu(image)
    else:
        print('No mask used.')
        mask = ones_like(image, dtype=bool)
else:
    try:
        mask = imread(args.mask).astype(bool)
    except FileNotFoundError:
        print('Cannot find mask file {}.'.format(args.mask))
        exit(1)
    print('Using mask file {}.'.format(args.mask))

if args.restrict:
    print('Restricting detection to the {}% brightest pixels.'.format(
        args.restrict))
    potential_centers = image > percentile(image[mask], 100 - args.restrict)
else:
    potential_centers = None

if args.normalize_mean:
    print('Normalizing image mean.')
    image = image / image.mean()

print('SPADEing the image...')
result, particles = spade2d(image=image,
                            shapes_library=shapes_library,
                            threshold=args.THRESHOLD,
                            mask=mask,
                            data_binding_function=None,
                            potential_centers=potential_centers,
                            return_candidates_list=True)
print('{} particles detected.'.format(result.max()))
print('    Y    X  Shape_ID')
print(' ' + sub('[\[\]]', '', array_str(particles)))

imsave(args.output, result)
print('Results written to {}'.format(args.output))



from skimage.io import imread
from skimage.filters import threshold_otsu
from numpy import percentile

from spade.detection_3d import spade3d
from spade.shapes.library import potatoids_5x5__smallest_4_pix

from imsave3d import imsave3d_overlay

# Load the example image.
image = imread("example_3d.tif")

# Separate cell image from background, using by Otsu's thresholding method.
cell = image > threshold_otsu(image)

# Focus on brightest pixels only
potential_centers = image > percentile(image[cell], 90)

# Detect particles.
particles = spade3d(image3d=image,
                    shapes_library=potatoids_5x5__smallest_4_pix,
                    threshold=118,
                    potential_centers=potential_centers,
                    mask=cell)

# Save result image. Open it with FIJI or your favorite TIFF stack viewer!
imsave3d_overlay('/tmp/spade3d_result.tif', image, cell, particles)



"""
This module provides different data binding functions to be used with SPADE.
In order to be able to be used with SPADE, a data binding function must:
- take 2 3D numpy arrays containing NaN entries as input
- output a 1D numpy array of spade 'scores' corresponding to the comparison of
the non-NaN values of each of the pixels population along the axis 0.

Example:

Input 1: shapes_pixels = [[[np.nan, np.nan, np.nan],
                           [np.nan,     99, np.nan],
                           [np.nan,     99, np.nan]],
                          [[np.nan,     99, np.nan],
                           [    99,     99, 99],
                           [np.nan,     99, np.nan]]]
Input 2: rings_pixels = [[[np.nan, np.nan, np.nan, np.nan, np.nan],
                          [np.nan,     44,     44,     44, np.nan],
                          [np.nan,     44, np.nan,     44, np.nan],
                          [np.nan,     44, np.nan,     44, np.nan],
                          [np.nan,     44,     44,     44, np.nan]],
                         [[np.nan,     44,     44,     44, np.nan],
                          [    44,     44, np.nan,     44,     44],
                          [    44, np.nan, np.nan, np.nan,     44],
                          [    44,     44, np.nan,     44,     44],
                          [np.nan,     44,     44,     44, np.nan]]]

Output: np.array([55, 55]) if the we expect a mean difference.
"""

import numpy as np
try:
    from scipy.stats import ttest_ind_from_stats as welch_test
except ImportError:
    pass  # TODO: implement a welch_test that doesn't require scipy.stats


def mean_difference(shapes_pixels, rings_pixels):
    return np.nanmean(shapes_pixels, axis=(1, 2)) - np.nanmean(rings_pixels,
                                                               axis=(1, 2))


def median_difference(shapes_pixels, rings_pixels):
    return np.nanmedian(shapes_pixels, axis=(1, 2)) - \
           np.nanmedian(rings_pixels, axis=(1, 2))


def max_min_difference(shapes_pixels, rings_pixels):
    return np.nanmin(shapes_pixels, axis=(1, 2)) - np.nanmax(rings_pixels,
                                                             axis=(1, 2))


# TODO: remove redundant calculations for these 3 functions
def t_test(shapes_pixels, rings_pixels):
    shapes_pixels_means = np.nanmean(shapes_pixels, axis=(1, 2))
    rings_pixels_means = np.nanmean(rings_pixels, axis=(1, 2))
    shapes_pixels_std = np.nanstd(shapes_pixels, axis=(1, 2))
    rings_pixels_std = np.nanstd(rings_pixels, axis=(1, 2))
    shapes_pixels_nobs = (~np.isnan(shapes_pixels)).sum(axis=(1, 2))
    rings_pixels_nobs = (~np.isnan(rings_pixels)).sum(axis=(1, 2))
    result = 1 - welch_test(shapes_pixels_means,
                            shapes_pixels_std,
                            shapes_pixels_nobs,
                            rings_pixels_means,
                            rings_pixels_std,
                            rings_pixels_nobs,
                            equal_var=False)[1]
    result[shapes_pixels_means - rings_pixels_means < 0] = -1
    return result


def bhattacharyya_distance(shapes_pixels, rings_pixels):
    shapes_pixels_variance = np.nanvar(shapes_pixels, axis=(1, 2))
    rings_pixels_variance = np.nanvar(rings_pixels, axis=(1, 2))
    shapes_pixels_variance[shapes_pixels_variance == 0] = 1
    rings_pixels_variance[rings_pixels_variance == 0] = 1
    mean_differences = mean_difference(shapes_pixels, rings_pixels)
    mean_differences[mean_differences <= 0] = np.nan
    result = -np.exp(-(1 / 4 * np.log(1 / 4 * (shapes_pixels_variance /
                                               rings_pixels_variance +
                                               rings_pixels_variance /
                                               shapes_pixels_variance + 2)) +
                       1 / 4 * (mean_differences ** 2) /
                       (shapes_pixels_variance + rings_pixels_variance))) + 1
    return result


def pseudo_bhattacharyya_distance(shapes_pixels, rings_pixels):
    shapes_pixels_variance = np.nanvar(shapes_pixels, axis=(1, 2))
    rings_pixels_variance = np.nanvar(rings_pixels, axis=(1, 2))
    shapes_pixels_variance[shapes_pixels_variance == 0] = 0.1
    rings_pixels_variance[rings_pixels_variance == 0] = 0.1
    mean_differences = mean_difference(shapes_pixels, rings_pixels)
    mean_differences[mean_differences <= 0] = np.nan
    result = -np.exp(-(1 / 4 * np.log(1 / 4 * (shapes_pixels_variance /
                                               rings_pixels_variance +
                                               rings_pixels_variance /
                                               shapes_pixels_variance + 2)) +
                       1 / 4 * (mean_differences ** 2) /
                       (np.sqrt(shapes_pixels_variance +
                                rings_pixels_variance)))) + 1
    return result



from distutils.core import setup
from distutils.command.install import INSTALL_SCHEMES

for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

setup(
    name='small-particle-detection',
    version='0.0.1a8',
    packages=['spade', 'spade.shapes'],
    url='http://spade.gforge.inria.fr/',
    license='CeCILL',
    author='Nicolas Cedilnik',
    author_email='nicolas.cedilnik@inria.fr',
    description='Small Particle Detection, a shapes library-based object '
                'detector.',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, '
        'version 2.1 (CeCILL-2.1)',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='image, image processing, scientific research',
    install_requires=['numpy'],
    # package_data={'': ['spade/shapes/*.npz', 'example/*.png']},
    include_package_data=True
)



"""
This file provides the ShapesLibrary class needed for SPADE, along with some
ready-to-use shapes libraries.
"""

from os.path import dirname, join
import warnings

import numpy as np
from scipy.ndimage.morphology import binary_dilation
from scipy.ndimage import generate_binary_structure


class ShapesLibrary(np.ndarray):
    """Shapes library class for SPADE"""

    # TODO: make size restriction easier. How? I don't know.
    def __new__(cls, input_array, name="Unknown", sort=True):
        if sort:
            input_array = np.array(sorted(input_array, key=np.sum), bool)
        elif type(input_array) != np.ndarray:
            input_array = np.asarray(input_array, bool)

        if input_array.ndim != 3 or input_array.shape[1] != \
                input_array.shape[2]:
            raise TypeError('Input array must be in the shape '
                            'number of shapes x grid_size x grid_size')

        obj = input_array.view(cls)
        obj.name = name
        obj.surfaces = get_surfaces(input_array)
        obj.rings = get_rings(input_array)
        obj.grid_size = input_array.shape[-1]
        obj.shape_start = start_dict(obj.surfaces)

        obj.half_shape_size = obj.grid_size // 2
        obj.image_extension_size = obj.half_shape_size + 1

        if not input_array[:, obj.half_shape_size, obj.half_shape_size].all():
            warnings.warn("At least one shape doesn't include the central "
                          "grid pixel. This may cause weird behaviour.")

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.name = getattr(obj, 'name', None)
        self.surfaces = getattr(obj, 'surfaces', None)
        self.rings = getattr(obj, 'rings', None)
        self.grid_size = getattr(obj, 'grid_size', None)
        self.shape_start = getattr(obj, 'shape_start', None)
        self.image_extension_size = getattr(obj, 'image_extension_size', None)
        self.half_shape_size = getattr(obj, 'half_shape_size', None)

        # def update(self):
        #     shapes_array = np.asarray(self)
        #     self.surfaces = get_surfaces(shapes_array)
        #     self.rings = get_rings(shapes_array)
        #     self.shape_start = start_dict(self.surfaces)


def get_surfaces(array):
    return array.sum(axis=tuple(np.arange(1, array.ndim)))


def start_dict(array):
    return {k: np.searchsorted(array, k) for k in np.unique(array)}


def get_rings(array, connectivity=None):
    if not connectivity:
        connectivity = array.ndim
    structuring_element = generate_binary_structure(array.ndim, connectivity)
    structuring_element[[0, -1]] = 0
    padded_array = np.pad(array, 1, 'constant')[1:-1]
    rings = np.logical_xor(
        binary_dilation(padded_array, structuring_element), padded_array)
    return rings


local_dir = dirname(__file__)
potatoids_5x5_array = np.load(join(local_dir,
                                   'potatoids_5x5.npz'))['arr_0.npy']
potatoids_5x5__smallest_1_pix = ShapesLibrary(potatoids_5x5_array,
                                              name='Potatoids 5x5, 1 to 25px',
                                              sort=False)
potatoids_5x5__smallest_2_pix = ShapesLibrary(potatoids_5x5_array[1:],
                                              name='Potatoids 5x5, 2 to 25px',
                                              sort=False)
potatoids_5x5__smallest_3_pix = ShapesLibrary(potatoids_5x5_array[3:],
                                              name='Potatoids 5x5, 3 to 25px',
                                              sort=False)
potatoids_5x5__smallest_4_pix = ShapesLibrary(potatoids_5x5_array[7:],
                                              name='Potatoids 5x5, 4 to 25px',
                                              sort=False)

potatoids_3x3_array = np.load(join(local_dir,
                                   'potatoids_3x3.npz'))['arr_0.npy']
potatoids_3x3__smallest_1_pix = ShapesLibrary(potatoids_3x3_array,
                                              name='Potatoids 3x3, 1 to 9px',
                                              sort=False)
potatoids_3x3__smallest_2_pix = ShapesLibrary(potatoids_3x3_array[1:],
                                              name='Potatoids 3x3, 2 to 9px',
                                              sort=False)
potatoids_3x3__smallest_3_pix = ShapesLibrary(potatoids_3x3_array[3:],
                                              name='Potatoids 3x3, 3 to 9px',
                                              sort=False)
potatoids_3x3__smallest_4_pix = ShapesLibrary(potatoids_3x3_array[7:],
                                              name='Potatoids 3x3, 4 to 9px',
                                              sort=False)

potatoids_5x5 = potatoids_5x5__smallest_3_pix
potatoids_3x3 = potatoids_3x3__smallest_3_pix



"""
This module provides functions to detect particles using SPADE on each slice
of a 3D image stack, and then combining them into 3D particles.
"""

import numpy as np
from .detection_2d import spade2d


def spade3d(image3d, shapes_library, threshold,
            mask=None,
            potential_centers=None,
            data_binding_function=None,
            minimal_surface_3ds=1,
            minimal_z_thickness=1,
            minimal_z_overlap=0.25,
            max_candidates=10000):
    # TODO: docstring
    candidates = np.zeros((max_candidates, 5), np.int16)
    candidates_idx = 0
    labeled_image_3d = np.zeros_like(image3d)
    if mask is None:
        mask = [None] * image3d.shape[0]
    if potential_centers is None:
        potential_centers = mask
    for z, image2d in enumerate(image3d):
        labeled_image_3d[z, :, :], tmp_candidates = \
            spade2d(image2d,
                    shapes_library,
                    threshold,
                    data_binding_function=data_binding_function,
                    mask=mask[z],
                    potential_centers=potential_centers[z],
                    return_candidates_list=True,
                    object_idx_start=candidates_idx + 1)
        candidates[candidates_idx: candidates_idx +
                                   len(tmp_candidates), 1:] = tmp_candidates
        candidates[candidates_idx: candidates_idx + len(tmp_candidates), 0] = z
        candidates_idx += len(tmp_candidates)
    candidates = candidates[:candidates_idx]
    return combine_2d_shapes(labeled_image_3d, candidates,
                             shapes_library, minimal_surface_3ds,
                             minimal_z_thickness, minimal_z_overlap)


def combine_2d_shapes(labmat, candidates, shapes, minsurf, minz, ztol):
    # now candidates are [z, y, x, shapeid, objectid], sorted by z, objectid is
    # the id in the labelmatrix
    finallabelmatrix = np.zeros(labmat.shape, dtype=np.uint8)
    zmax = labmat.shape[0]
    nbcand = len(candidates)
    i = 0
    objid = 1
    exclude = []
    potobj = np.zeros(zmax, dtype=np.uint16)
    while i < nbcand:
        if candidates[i, 4] not in exclude:
            z1, y1, x1, shapeid1, oid1 = candidates[i, :]
            potobjlen = 1
            potobj[0] = oid1
            sameobject = True
            targz = z1 + 1
            j = i + 1
            while sameobject:
                sameobject = False
                while j < nbcand:
                    if candidates[j][0] == z1:
                        j += 1
                    elif candidates[j][0] > targz:
                        break
                    else:
                        z2, y2, x2, shapeid2, oid2 = candidates[j, :]
                        if overlap(labmat, z1, z2, oid1, oid2) / \
                                shapes.surfaces[min(shapeid1, shapeid2)] >= \
                                ztol:
                            potobj[potobjlen] = oid2
                            potobjlen += 1
                            sameobject = True
                            z1, y1, x1, shapeid1, oid1 = z2, y2, x2, \
                                                         shapeid2, oid2
                            targz = z1 + 1
                            j += 1
                            break
                        j += 1
            if potobjlen >= minz:
                valid = False
                for sid in candidates[potobj[:potobjlen] - 1, 3]:
                    if shapes.surfaces[sid] >= minsurf:
                        valid = True
                        break
                if valid:
                    rpotobj = potobj[:potobjlen].tolist()
                    exclude += rpotobj
                    splitted_obj = np.logical_or.reduce([labmat == x for x
                                                         in rpotobj])
                    finallabelmatrix[splitted_obj] = objid
                    objid += 1
        i += 1
    return finallabelmatrix


def overlap(labmat, z1, z2, oid1, oid2):
    loclabmat = labmat[z1:z2 + 1, :, :]
    loclabmat = np.where(np.logical_or(loclabmat == oid1,
                                       loclabmat == oid2), True, False)
    return np.logical_and.reduce(loclabmat, 0).sum()



"""
This file was used to generate potatoids_5x5 and potatoids_3x3
You shouldn't need it!
"""
# TODO: still a lot of cleaning

import numpy as np
from scipy import ndimage


# from spade.shapes.library import ShapesLibrary


def intervalstoboolarray(intervals, grid_size):
    # generates an array from a list of intervals
    array = np.zeros((grid_size + 2, grid_size + 2), dtype=np.bool)
    for i, interval in enumerate(intervals):
        if interval:
            start = interval[0]
            end = interval[1]
            array[i + 1] = [False] * (start + 1) + (end - start + 1) * [
                True] + (grid_size - end) * [False]
        else:
            array[i + 1] = [False] * (grid_size + 2)
    return array


def finalarray(shapes, sepcolor, septhickness, outputwidth):
    print('Creating the big array...')
    grid_size = np.shape(shapes[0])[0]
    shapesbyline = outputwidth // (grid_size + 2 + septhickness)
    outputwidth = shapesbyline * (grid_size + 2 + septhickness) - 1
    sep = np.reshape(
        np.array([[sepcolor] * grid_size * septhickness], dtype=np.uint8),
        (grid_size, septhickness))
    seph = np.array([sepcolor] * outputwidth, dtype=np.uint8)
    res = seph
    for i in range(0, len(shapes), shapesbyline):
        rest = shapes[i]
        for forme in shapes[i + 1:i + shapesbyline]:
            rest = np.concatenate([rest, sep, forme], axis=1)
        while rest.shape[1] < len(seph):
            rest = np.concatenate([rest, sep], axis=1)
        res = np.vstack([res, rest, seph])
    return res


def valid_line_configurations(prev_conf, grid_size, side, max_shift):
    # given a line, returns the possible 'following' lines
    res = [()]
    if prev_conf:
        prev_start = prev_conf[0]
        prev_end = prev_conf[1]
        res += [(prev_start + x, prev_end - y)
                for x in range(maximum_shift + 1)
                for y in range(maximum_shift + 1)
                if prev_end - y >= prev_start + x]
        prev_length = prev_end - prev_start + 1
        if prev_length > max_shift:
            if side == right:
                for shift in range(1, max_shift + 1):
                    max_start = min(prev_start + shift, grid_size - 1)
                    max_end = min(prev_end + shift, grid_size - 1)
                    if max_end >= max_start:
                        res += [(max_start, max_end)]
                    if prev_end >= max_start:
                        res += [(max_start, prev_end)]
            elif side == left:
                for shift in range(1, max_shift + 1):
                    min_start = max(prev_start - shift, 0)
                    min_end = max(prev_end - shift, 0)
                    if min_end >= min_start:
                        res += [(min_start, min_end)]
                    if min_end >= prev_start:
                        res += [(prev_start, min_end)]
    return res


def shapesize(config):
    res = 0
    for line in config:
        if line:
            res += line[1] - line[0] + 1
    return res


def uniques(shapes, grid_size):
    shapesdef = tuple()
    while shapes:
        shape = shapes.pop(0)
        shapesdef += shape,
        r = list(range(-(grid_size // 2), 0)) + list(
            range(1, (grid_size // 2) + 1))
        identicalshapes = tuple()
        # left/right translations
        for i in r:
            ident = tuple()
            for j in shape:
                if j:
                    ident += (j[0] + i, j[1] + i),
                else:
                    ident += (),
            identicalshapes += (ident),
        # up/down translations
        emptylinesabove = 0
        while shape[emptylinesabove] == ():
            emptylinesabove += 1
        for ishape in ((shape,) + identicalshapes):
            for i in range(0, emptylinesabove):
                ident = list(ishape[i + 1:]) + [()] * (i + 1)
                identicalshapes += (tuple(ident)),
        i = grid_size - 1
        emptylinesbelow = 0
        while shape[i] == ():
            emptylinesbelow += 1
            i -= 1
        for ishape in ((shape,) + identicalshapes):
            for i in range(0, emptylinesbelow):
                ident = [()] * (i + 1) + list(ishape[:(grid_size - i - 1)])
                identicalshapes += (tuple(ident)),
        for s2 in identicalshapes:
            if s2 in shapes: shapes.remove(s2)
    return shapesdef


def centershape(shape, target):
    row, col = ndimage.measurements.center_of_mass(shape)
    row = int(round(row))
    col = int(round(col))
    shape = np.roll(np.roll(shape, target - row, 0), target - col)
    if shape[:, 0].any():
        shape = np.roll(shape, 1)
    if shape[:, -1].any():
        shape = np.roll(shape, -1)
    if shape[0, :].any():
        shape = np.roll(shape, 1, 0)
    if shape[-1, :].any():
        shape = np.roll(shape, -1, 0)
    if shape[:, 0].any() or shape[:, -1].any() or \
            shape[0, :].any() or shape[-1, :].any():
        print('Shape cannot be centered.')
    return shape


def notlinear(shapes):
    shapesdef = tuple()
    while shapes:
        shape = shapes.pop(0)
        vlinear = True
        filledlines = grid_size
        for line in shape:
            if line:
                if line[1] - line[0] > 0:
                    vlinear = False
            else:
                filledlines -= 1
        if not vlinear and filledlines > 1: shapesdef += shape,
    return shapesdef


def isshift(l1, l2):
    s1 = l1[0]
    s2 = l2[0]
    e1 = l1[1]
    e2 = l2[1]
    if s2 < s1 or e2 > e1: return True
    return False


def generateshapes(grid_size, minsize, maxsize, max_shift):
    center = grid_size // 2  # origin of shape
    centrallines = [[(x, y)] for x in range(grid_size)
                    for y in range(grid_size)
                    if x <= center <= y]  # only lines touching the origin
    configs = []
    j = 0
    print('Computing lines...')
    for side in (left, right):
        configstmp = centrallines[:]  # starting point
        for i in range(center):
            j += 1
            extended_configs = []
            print('%i/%i...' % (j, center * 2))
            # extend from the central line to the bottom
            for config in configstmp:
                if len(config) > 1 and config[-1] and isshift(config[-2],
                                                              config[-1]):
                    shift = max_shift
                else:
                    shift = 0
                next_line_configs = valid_line_configurations(config[-1],
                                                              grid_size,
                                                              side, shift)
                extended_configs.extend([config + [next_line_config]
                                         for next_line_config in
                                         next_line_configs])
            configstmp = extended_configs
            # extend from the bottom half-shape to the top
            extended_configs = []
            for config in configstmp:
                if config[0] and config[1] and isshift(config[1], config[0]):
                    shift = max_shift
                else:
                    shift = 0
                next_line_configs = valid_line_configurations(config[0],
                                                              grid_size,
                                                              not side, shift)
                extended_configs.extend([[next_line_config] + config
                                         for next_line_config in
                                         next_line_configs])
            configstmp = extended_configs
        configs += configstmp
    print('%i shapes' % len(configs))
    print('Removing shapes based on size...')
    shapes = [tuple(config) for config in configs
              if minsize <= shapesize(config) <= maxsize]
    print('%i shapes' % len(shapes))
    print('Removing linear shapes...')
    shapes = notlinear(shapes)
    print('%i shapes' % len(shapes))
    print('Removing duplicates...')
    shapes = list(set(shapes))
    print('%i shapes' % len(shapes))
    print('Removing duplicates by translation...')
    shapes = uniques(shapes, grid_size)
    print('%i shapes' % len(shapes))
    print('Converting lists of intervals to boolean arrays...')
    res = [intervalstoboolarray(config, grid_size) for config in shapes]
    print('Centering shapes...')
    res = [centershape(shape, center + 1) for shape in res]
    print('Sorting arrays...')
    res.sort(key=np.sum)
    return res


def com(shape):
    row, col = ndimage.measurements.center_of_mass(shape)
    row = grid_size // 2 - int(round(row))
    col = grid_size // 2 - int(round(col))
    return row, col


# shape parameters
grid_size = 3  # edge of square containing shapes
minimal_shape_surface = 1  # minimum size of object
maximal_shape_size = 9  # minimum size of object
maximum_shift = 1

# rendering parameters
bgcolor = 255  # background color
color = 0  # object color
sepcolor = 200  # separator color
septhickness = 1  # separator thickness
outputwidth = 150  # max width of image output

###############################################################################
right = True
left = False
conn8 = ndimage.generate_binary_structure(2, 2)

# generate shapes
shapes = generateshapes(grid_size, minimal_shape_surface, maximal_shape_size,
                        maximum_shift)

# add 1 pix and 2 pix shapes
p1 = np.zeros((grid_size + 2, grid_size + 2), bool)
p1[grid_size // 2 + 1, grid_size // 2 + 1] = True
p2 = np.zeros((grid_size + 2, grid_size + 2), bool)
p2[grid_size // 2 + 1:grid_size // 2 + 3, grid_size // 2 + 1] = True
p3 = np.zeros((grid_size + 2, grid_size + 2), bool)
p3[grid_size // 2 + 1, grid_size // 2 + 1:grid_size // 2 + 3] = True
shapes = [p1, p2, p3] + shapes

shapesarray = np.array([s[1:-1, 1:-1] for s in shapes])

# potrnps = ShapesLibrary(shapesarray, name="Patatoids 5x5")

np.savez("patatoids_3x3", shapesarray)
